#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <stdlib.h>

uint8_t sum;
uint8_t *inbf, *outbf;
uint32_t filesize;

int terminal_error(void){
  if(inbf)
    free(inbf);
  if(outbf)
    free(outbf);
  exit(-1);
}

uint32_t file_save(const char * filename, void * buf, uint32_t len)
{
	FILE * out = fopen(filename, "wb");
	int r;
	if(!out)
	{
    return 0;
	}
	r = fwrite(buf, len, 1, out);
	fclose(out);
	return r;
}

void * file_load(const char * filename, uint32_t * len)
{
  uint32_t size, n;
  FILE * in;
  char * buf;
  in = fopen(filename, "rb");
  fseek(in, 0, SEEK_END);                                   // seek to end of file
  size = ftell(in);                                         // get current file pointer
  fseek(in, 0, SEEK_SET);                                   // seek back to beginning of file
	buf = malloc(size);                                       // allocate size
  if(!buf){
    printf("Unable to allocate file buffer!\r\n");
    return NULL;
  }
  n = fread(buf, 1, size, in);                              // Read whole file

  if(n < size){                                             // Ensure it was completely read
      free(buf);
      return NULL;
  }

	if(len)
		*len = n;

	if(in != stdin)
    fclose(in);
  return buf;
}


void usage(void)
{
  printf("\r\n  tool [input] [output] [xor]\r\n\n");
  printf("  This tool is compatible with the .AG firmware file of Agilent/Keysight U1253B, U1273A(X) and maybe others.\r\n");
  printf("  It reverses the byte order and swaps the nibbles, optionally applying a XOR value.\r\n");
  printf("  The output will be the original firmware bytes, ready for modifying or disassembling.\r\n");
  printf("  Once modified, run the tool again to convert it back to the original format, it'll be ready for flashing.\r\n\n");
  printf("  XOR option is optional and only for research purposes, not required to reverse the original .AG firmware file.\r\n");
  printf("  XOR value must written in hex representation, like 0x9A.\r\n");
  printf("  Examples: tool input.bin output.bin\r\n");
  printf("            tool input.bin output.bin 0x9A\r\n\n");
  printf("  More info at https://www.eevblog.com/forum/microcontrollers/8-bit-78k0-d78f0547-help-with-firmware-modification/\r\n\n");
}

int main(int argc, char * argv[])
{
  if(argc < 3) {
    usage();
    exit(-1);
  }

  inbf = file_load(argv[1], &filesize);
  if(!inbf){
    printf("Unable to read from file \"%s\"!\r\n", argv[1]);
    terminal_error();
  }
  printf("Read file OK: %s, %lu bytes\r\n", argv[1], filesize);

  outbf = malloc(filesize);
  if(!outbf){
    printf("Unable to allocate output buffer!\r\n");
    terminal_error();
  }

  sum = strtol(argv[3], NULL, 16);

  for(uint32_t i=0; i<filesize; i++){
    outbf[i] = inbf[filesize-1-i];
    outbf[i] = ((outbf[i] & 0xF0) >>4) | ((outbf[i] & 0x0F) <<4);
    outbf[i] ^= sum;
  }

  if(!file_save(argv[2], outbf, filesize)){
    printf("Unable to write to file \"%s\"!\r\n", argv[2]);
    terminal_error();
  }

  free(inbf);
  free(outbf);

  printf("%lu bytes saved to file \"%s\"\r\n", filesize, argv[2]);
  printf("Done!\r\n");
  return 0;
}

